import mongoengine
import pymongo
import pytest
from mongoengine.connection import ConnectionFailure
from mongoengine.context_managers import switch_db
from pymongo.database import Database
from pymongo.errors import InvalidURI
from pymongo.mongo_client import MongoClient
from pymongo.read_preferences import ReadPreference

from flask_mongoengine import MongoEngine, current_mongoengine_instance


def test_connection__should_use_defaults__if_no_settings_provided(app):
    """Make sure a simple connection to a standalone MongoDB works."""
    db = MongoEngine()

    # Verify no extension for Mongoengine yet created for app
    assert app.extensions == {}
    assert current_mongoengine_instance() is None

    # Create db connection. Should return None.
    assert db.init_app(app) is None

    # Verify db added to Flask extensions.
    assert current_mongoengine_instance() == db

    # Verify db settings passed to pymongo driver.
    # Default mongoengine db is 'default', default Flask-Mongoengine db is 'test'.
    connection = mongoengine.get_connection()
    mongo_engine_db = mongoengine.get_db()
    assert isinstance(mongo_engine_db, Database)
    assert isinstance(connection, MongoClient)
    assert mongo_engine_db.name == "test"
    assert connection.HOST == "localhost"
    assert connection.PORT == 27017


@pytest.mark.parametrize(
    ("config_extension"),
    [
        {
            "MONGODB_SETTINGS": {
                "ALIAS": "simple_conn",
                "HOST": "localhost",
                "PORT": 27017,
                "DB": "flask_mongoengine_test_db",
            }
        },
        {
            "MONGODB_HOST": "localhost",
            "MONGODB_PORT": 27017,
            "MONGODB_DB": "flask_mongoengine_test_db",
            "MONGODB_ALIAS": "simple_conn",
        },
    ],
    ids=("Dict format", "Config variable format"),
)
def test_connection__should_pass_alias__if_provided(app, config_extension):
    """Make sure a simple connection pass ALIAS setting variable."""
    db = MongoEngine()
    app.config.update(config_extension)

    # Verify no extension for Mongoengine yet created for app
    assert app.extensions == {}
    assert current_mongoengine_instance() is None

    # Create db connection. Should return None.
    assert db.init_app(app) is None

    # Verify db added to Flask extensions.
    assert current_mongoengine_instance() == db

    # Verify db settings passed to pymongo driver.
    # ALIAS is used to find correct connection.
    # As we do not use default alias, default call to mongoengine.get_connection
    # should raise.
    with pytest.raises(ConnectionFailure):
        mongoengine.get_connection()

    connection = mongoengine.get_connection("simple_conn")
    mongo_engine_db = mongoengine.get_db("simple_conn")
    assert isinstance(mongo_engine_db, Database)
    assert isinstance(connection, MongoClient)
    assert mongo_engine_db.name == "flask_mongoengine_test_db"
    assert connection.HOST == "localhost"
    assert connection.PORT == 27017


@pytest.mark.parametrize(
    ("config_extension"),
    [
        {
            "MONGODB_SETTINGS": {
                "HOST": "mongodb://localhost:27017/flask_mongoengine_test_db"
            }
        },
        {
            "MONGODB_HOST": "mongodb://localhost:27017/flask_mongoengine_test_db",
            "MONGODB_PORT": 27017,
            "MONGODB_DB": "should_ignore_it",
        },
    ],
    ids=("Dict format", "Config variable format"),
)
def test_connection__should_parse_host_uri__if_host_formatted_as_uri(
    app, config_extension
):
    """Make sure a simple connection pass ALIAS setting variable."""
    db = MongoEngine()
    app.config.update(config_extension)

    # Verify no extension for Mongoengine yet created for app
    assert app.extensions == {}
    assert current_mongoengine_instance() is None

    # Create db connection. Should return None.
    assert db.init_app(app) is None

    # Verify db added to Flask extensions.
    assert current_mongoengine_instance() == db

    connection = mongoengine.get_connection()
    mongo_engine_db = mongoengine.get_db()
    assert isinstance(mongo_engine_db, Database)
    assert isinstance(connection, MongoClient)
    assert mongo_engine_db.name == "flask_mongoengine_test_db"
    assert connection.HOST == "localhost"
    assert connection.PORT == 27017


@pytest.mark.parametrize(
    ("config_extension"),
    [
        {
            "MONGODB_SETTINGS": {
                "HOST": "mongomock://localhost:27017/flask_mongoengine_test_db"
            }
        },
        {
            "MONGODB_SETTINGS": {
                "ALIAS": "simple_conn",
                "HOST": "localhost",
                "PORT": 27017,
                "DB": "flask_mongoengine_test_db",
                "IS_MOCK": True,
            }
        },
        {"MONGODB_HOST": "mongomock://localhost:27017/flask_mongoengine_test_db"},
    ],
    ids=("Dict format as URI", "Dict format as Param", "Config variable format as URI"),
)
def test_connection__should_parse_mongo_mock_uri__as_uri_and_as_settings(
    app, config_extension
):
    """Make sure a simple connection pass ALIAS setting variable."""
    db = MongoEngine()
    app.config.update(config_extension)

    # Verify no extension for Mongoengine yet created for app
    assert app.extensions == {}
    assert current_mongoengine_instance() is None

    # Create db connection. Should return None.

    with pytest.raises(RuntimeError) as error:
        assert db.init_app(app) is None

    assert str(error.value) == "You need mongomock installed to mock MongoEngine."


@pytest.mark.parametrize(
    ("config_extension"),
    [
        {
            "MONGODB_SETTINGS": {
                "HOST": "postgre://localhost:27017/flask_mongoengine_test_db"
            }
        },
        {"MONGODB_HOST": "mysql://localhost:27017/flask_mongoengine_test_db"},
    ],
    ids=("Dict format as URI", "Config variable format as URI"),
)
def test_connection__should_raise__if_uri_not_properly_formatted(app, config_extension):
    """Make sure a simple connection pass ALIAS setting variable."""
    db = MongoEngine()
    app.config.update(config_extension)

    # Verify no extension for Mongoengine yet created for app
    assert app.extensions == {}
    assert current_mongoengine_instance() is None

    # Create db connection. Should return None.

    with pytest.raises(InvalidURI) as error:
        assert db.init_app(app) is None

    assert (
        str(error.value)
        == "Invalid URI scheme: URI must begin with 'mongodb://' or 'mongodb+srv://'"
    )


def test_connection__should_accept_host_as_list(app):
    """Make sure MONGODB_HOST can be a list hosts."""
    db = MongoEngine()
    app.config["MONGODB_SETTINGS"] = {
        "ALIAS": "host_list",
        "HOST": ["localhost:27017"],
        "DB": "flask_mongoengine_list_test_db",
    }
    db.init_app(app)

    connection = mongoengine.get_connection("host_list")
    mongo_engine_db = mongoengine.get_db("host_list")
    assert isinstance(mongo_engine_db, Database)
    assert isinstance(connection, MongoClient)
    assert mongo_engine_db.name == "flask_mongoengine_list_test_db"
    assert connection.HOST == "localhost"
    assert connection.PORT == 27017


def test_multiple_connections(app):
    """Make sure establishing multiple connections to a standalone
    MongoDB and switching between them works.
    """
    db = MongoEngine()
    app.config["MONGODB_SETTINGS"] = [
        {
            "ALIAS": "default",
            "DB": "flask_mongoengine_test_db_1",
            "HOST": "localhost",
            "PORT": 27017,
        },
        {
            "ALIAS": "alternative",
            "DB": "flask_mongoengine_test_db_2",
            "HOST": "localhost",
            "PORT": 27017,
        },
    ]

    class Todo(db.Document):
        title = db.StringField(max_length=60)

    db.init_app(app)
    # Drop default collection from init
    Todo.drop_collection()
    Todo.meta = {"db_alias": "alternative"}
    # Drop 'alternative' collection initiated early.
    Todo.drop_collection()

    # Make sure init correct and both databases are clean
    with switch_db(Todo, "default") as Todo:
        doc = Todo.objects().first()
        assert doc is None

    with switch_db(Todo, "alternative") as Todo:
        doc = Todo.objects().first()
        assert doc is None

    # Test saving a doc via the default connection
    with switch_db(Todo, "default") as Todo:
        todo = Todo()
        todo.text = "Sample"
        todo.title = "Testing"
        todo.done = True
        s_todo = todo.save()

        f_to = Todo.objects().first()
        assert s_todo.title == f_to.title

    # Make sure the doc still doesn't exist in the alternative db
    with switch_db(Todo, "alternative") as Todo:
        doc = Todo.objects().first()
        assert doc is None

    # Make sure switching back to the default connection shows the doc
    with switch_db(Todo, "default") as Todo:
        doc = Todo.objects().first()
        assert doc is not None


def test_ingnored_mongodb_prefix_config(app):
    """Config starting by MONGODB_ but not used by flask-mongoengine
    should be ignored.
    """
    db = MongoEngine()
    app.config[
        "MONGODB_HOST"
    ] = "mongodb://localhost:27017/flask_mongoengine_test_db_prod"
    # Invalid host, should trigger exception if used
    app.config["MONGODB_TEST_HOST"] = "dummy://localhost:27017/test"
    db.init_app(app)

    connection = mongoengine.get_connection()
    mongo_engine_db = mongoengine.get_db()
    assert isinstance(mongo_engine_db, Database)
    assert isinstance(connection, MongoClient)
    assert mongo_engine_db.name == "flask_mongoengine_test_db_prod"
    assert connection.HOST == "localhost"
    assert connection.PORT == 27017


def test_connection_kwargs(app):
    """Make sure additional connection kwargs work."""

    # Figure out whether to use "MAX_POOL_SIZE" or "MAXPOOLSIZE" based
    # on PyMongo version (former was changed to the latter as described
    # in https://jira.mongodb.org/browse/PYTHON-854)
    # TODO remove once PyMongo < 3.0 support is dropped
    if pymongo.version_tuple[0] >= 3:
        MAX_POOL_SIZE_KEY = "MAXPOOLSIZE"
    else:
        MAX_POOL_SIZE_KEY = "MAX_POOL_SIZE"

    app.config["MONGODB_SETTINGS"] = {
        "ALIAS": "tz_aware_true",
        "DB": "flask_mongoengine_testing_tz_aware",
        "TZ_AWARE": True,
        "READ_PREFERENCE": ReadPreference.SECONDARY,
        MAX_POOL_SIZE_KEY: 10,
    }
    db = MongoEngine(app)

    assert db.connection.codec_options.tz_aware
    assert db.connection.max_pool_size == 10
    assert db.connection.read_preference == ReadPreference.SECONDARY



import copy
import functools
import inspect
import os
import socketserver
import sys
import time

import bson
import pymongo.collection
import pymongo.command_cursor
import pymongo.cursor
import pymongo.helpers

__all__ = [
    "queries",
    "inserts",
    "updates",
    "removes",
    "install_tracker",
    "uninstall_tracker",
    "reset",
    "response_sizes",
]

_original_methods = {
    "insert": pymongo.collection.Collection.insert,
    "update": pymongo.collection.Collection.update,
    "remove": pymongo.collection.Collection.remove,
    "refresh": pymongo.cursor.Cursor._refresh,
    "_unpack_response": pymongo.command_cursor.CommandCursor._unpack_response,
}

queries = []
inserts = []
updates = []
removes = []
response_sizes = []

if sys.version_info >= (3, 0):
    unicode = str


# Wrap helpers._unpack_response for getting response size
@functools.wraps(_original_methods["_unpack_response"])
def _unpack_response(response, *args, **kwargs):
    result = _original_methods["_unpack_response"](response, *args, **kwargs)
    response_sizes.append(sys.getsizeof(response, len(response)) / 1024.0)
    return result


# Wrap Cursor.insert for getting queries
@functools.wraps(_original_methods["insert"])
def _insert(
    collection_self, doc_or_docs, manipulate=True, safe=None, check_keys=True, **kwargs
):
    start_time = time.time()
    result = _original_methods["insert"](
        collection_self, doc_or_docs, check_keys=check_keys, **kwargs
    )
    total_time = (time.time() - start_time) * 1000

    __traceback_hide__ = True  # noqa
    stack_trace, internal = _tidy_stacktrace()
    inserts.append(
        {
            "document": doc_or_docs,
            "time": total_time,
            "stack_trace": stack_trace,
            "size": response_sizes[-1] if response_sizes else 0,
            "internal": internal,
        }
    )
    return result


# Wrap Cursor.update for getting queries
@functools.wraps(_original_methods["update"])
def _update(
    collection_self,
    spec,
    document,
    upsert=False,
    maniuplate=False,
    safe=None,
    multi=False,
    **kwargs,
):
    start_time = time.time()
    result = _original_methods["update"](
        collection_self, spec, document, upsert=upsert, multi=multi, **kwargs
    )
    total_time = (time.time() - start_time) * 1000

    __traceback_hide__ = True  # noqa
    stack_trace, internal = _tidy_stacktrace()
    updates.append(
        {
            "document": document,
            "upsert": upsert,
            "multi": multi,
            "spec": spec,
            "time": total_time,
            "stack_trace": stack_trace,
            "size": response_sizes[-1] if response_sizes else 0,
            "internal": internal,
        }
    )
    return result


# Wrap Cursor.remove for getting queries
@functools.wraps(_original_methods["remove"])
def _remove(collection_self, spec_or_id, safe=None, **kwargs):
    start_time = time.time()
    result = _original_methods["remove"](collection_self, spec_or_id, **kwargs)
    total_time = (time.time() - start_time) * 1000

    __traceback_hide__ = True  # noqa
    stack_trace, internal = _tidy_stacktrace()
    removes.append(
        {
            "spec_or_id": spec_or_id,
            "time": total_time,
            "   ": stack_trace,
            "size": response_sizes[-1] if response_sizes else 0,
            "internal": internal,
        }
    )
    return result


# Wrap Cursor._refresh for getting queries
@functools.wraps(_original_methods["refresh"])
def _cursor_refresh(cursor_self):
    # Look up __ private instance variables
    def privar(name):
        return getattr(cursor_self, "_Cursor__{0}".format(name), None)

    if privar("id") is not None:
        # getMore not query - move on
        return _original_methods["refresh"](cursor_self)

    # NOTE: See pymongo/cursor.py+557 [_refresh()] and
    # pymongo/message.py for where information is stored

    # Time the actual query
    start_time = time.time()
    result = _original_methods["refresh"](cursor_self)
    total_time = (time.time() - start_time) * 1000

    query_son = privar("query_spec")()
    if not isinstance(query_son, bson.SON):

        if "$query" not in query_son:
            query_son = {"$query": query_son}

        data = privar("data")
        if data:
            query_son["data"] = data

        orderby = privar("ordering")
        if orderby:
            query_son["$orderby"] = orderby

        hint = privar("hint")
        if hint:
            query_son["$hint"] = hint

        snapshot = privar("snapshot")
        if snapshot:
            query_son["$snapshot"] = snapshot

        maxScan = privar("max_scan")
        if maxScan:
            query_son["$maxScan"] = maxScan

    __traceback_hide__ = True  # noqa
    stack_trace, internal = _tidy_stacktrace()
    query_data = {
        "time": total_time,
        "operation": "query",
        "stack_trace": stack_trace,
        "size": response_sizes[-1] if response_sizes else 0,
        "data": copy.copy(privar("data")),
        "internal": internal,
    }

    # Collection in format <db_name>.<collection_name>
    collection_name = privar("collection")
    query_data["collection"] = collection_name.full_name.split(".")[1]

    if query_data["collection"] == "$cmd":
        query_data["operation"] = "command"
        # Handle count as a special case
        if "count" in query_son:
            # Information is in a different format to a standard query
            query_data["collection"] = query_son["count"]
            query_data["operation"] = "count"
            query_data["skip"] = query_son.get("skip")
            query_data["limit"] = query_son.get("limit")
            query_data["query"] = query_son["query"]
    else:
        # Normal Query
        query_data["skip"] = privar("skip")
        query_data["limit"] = privar("limit")
        query_data["query"] = query_son["$query"]
        query_data["ordering"] = _get_ordering(query_son)

    queries.append(query_data)

    return result


def install_tracker():
    if pymongo.collection.Collection.insert != _insert:
        pymongo.collection.Collection.insert = _insert
    if pymongo.collection.Collection.update != _update:
        pymongo.collection.Collection.update = _update
    if pymongo.collection.Collection.remove != _remove:
        pymongo.collection.Collection.remove = _remove
    if pymongo.cursor.Cursor._refresh != _cursor_refresh:
        pymongo.cursor.Cursor._refresh = _cursor_refresh
    if pymongo.command_cursor.CommandCursor._unpack_response != _unpack_response:
        pymongo.command_cursor.CommandCursor._unpack_response = _unpack_response


def uninstall_tracker():
    if pymongo.collection.Collection.insert == _insert:
        pymongo.collection.Collection.insert = _original_methods["insert"]
    if pymongo.collection.Collection.update == _update:
        pymongo.collection.Collection.update = _original_methods["update"]
    if pymongo.collection.Collection.remove == _remove:
        pymongo.collection.Collection.remove = _original_methods["remove"]
    if pymongo.cursor.Cursor._refresh == _cursor_refresh:
        pymongo.cursor.Cursor._refresh = _original_methods["cursor_refresh"]
    if pymongo.command_cursor.CommandCursor._unpack_response == _unpack_response:
        pymongo.command_cursor.CommandCursor._unpack_response = _original_methods[
            "_unpack_response"
        ]


def reset():
    global queries, inserts, updates, removes, response_sizes
    queries = []
    inserts = []
    updates = []
    removes = []
    response_sizes = []


def _get_ordering(son):
    """Helper function to extract formatted ordering from dict."""

    def fmt(field, direction):
        return "{0}{1}".format({-1: "-", 1: "+"}[direction], field)

    if "$orderby" in son:
        return ", ".join(fmt(f, d) for f, d in son["$orderby"].items())


def _tidy_stacktrace():
    """
    Tidy the stack_trace
    """
    socketserver_path = os.path.realpath(os.path.dirname(socketserver.__file__))
    pymongo_path = os.path.realpath(os.path.dirname(pymongo.__file__))
    paths = ["/site-packages/", "/flaskext/", socketserver_path, pymongo_path]
    internal = False

    # Check html templates
    fnames = []
    for i in range(100):
        try:
            fname = sys._getframe(i).f_code.co_filename
            if ".html" in fname:
                fnames.append(fname)
        except Exception:
            break
    fnames = list(set(fnames))
    trace = []

    for path in fnames:
        if "flask_debugtoolbar" in path:
            internal = True
        trace.append((path, "?", "?", "?", False))

    if trace:
        return trace, internal

    stack = inspect.stack()
    reversed(stack)

    trace = []
    for frame, path, line_no, func_name, text in (f[:5] for f in stack):
        s_path = os.path.realpath(path)
        # Support hiding of frames -- used in various utilities that provide
        # inspection.
        if "__traceback_hide__" in frame.f_locals:
            continue
        hidden = False
        if func_name == "<genexpr>":
            hidden = True
        if any([p for p in paths if p in s_path]):
            hidden = True
        if not text:
            text = ""
        else:
            if sys.version_info >= (3, 0):
                text = "".join(text).strip()
            else:
                try:
                    text = unicode("".join(text).strip())
                except Exception:
                    pass
        trace.append((path, line_no, func_name, text, hidden))
    return trace, internal



"""
Useful form fields for use with the mongoengine.
"""
import json
import sys
from gettext import gettext as _

from mongoengine.queryset import DoesNotExist
from wtforms import widgets
from wtforms.fields import SelectFieldBase, StringField, TextAreaField
from wtforms.validators import ValidationError

__all__ = (
    "ModelSelectField",
    "QuerySetSelectField",
)

if sys.version_info >= (3, 0):
    unicode = str


class QuerySetSelectField(SelectFieldBase):
    """
    Given a QuerySet either at initialization or inside a view, will display a
    select drop-down field of choices. The `data` property actually will
    store/keep an ORM model instance, not the ID. Submitting a choice which is
    not in the queryset will result in a validation error.

    Specifying `label_attr` in the constructor will use that property of the
    model instance for display in the list, else the model object's `__str__`
    or `__unicode__` will be used.

    If `allow_blank` is set to `True`, then a blank choice will be added to the
    top of the list. Selecting this choice will result in the `data` property
    being `None`.  The label for the blank choice can be set by specifying the
    `blank_text` parameter.
    """

    widget = widgets.Select()

    def __init__(
        self,
        label="",
        validators=None,
        queryset=None,
        label_attr="",
        allow_blank=False,
        blank_text="---",
        label_modifier=None,
        **kwargs,
    ):

        super(QuerySetSelectField, self).__init__(label, validators, **kwargs)
        self.label_attr = label_attr
        self.allow_blank = allow_blank
        self.blank_text = blank_text
        self.label_modifier = label_modifier
        self.queryset = queryset

    def iter_choices(self):
        if self.allow_blank:
            yield ("__None", self.blank_text, self.data is None)

        if self.queryset is None:
            return

        self.queryset.rewind()
        for obj in self.queryset:
            label = (
                self.label_modifier(obj)
                if self.label_modifier
                else (self.label_attr and getattr(obj, self.label_attr) or obj)
            )

            if isinstance(self.data, list):
                selected = obj in self.data
            else:
                selected = self._is_selected(obj)
            yield (obj.id, label, selected)

    def process_formdata(self, valuelist):
        if valuelist:
            if valuelist[0] == "__None":
                self.data = None
            else:
                if self.queryset is None:
                    self.data = None
                    return

                try:
                    obj = self.queryset.get(pk=valuelist[0])
                    self.data = obj
                except DoesNotExist:
                    self.data = None

    def pre_validate(self, form):
        if not self.allow_blank or self.data is not None:
            if not self.data:
                raise ValidationError(_("Not a valid choice"))

    def _is_selected(self, item):
        return item == self.data


class QuerySetSelectMultipleField(QuerySetSelectField):

    widget = widgets.Select(multiple=True)

    def __init__(
        self,
        label="",
        validators=None,
        queryset=None,
        label_attr="",
        allow_blank=False,
        blank_text="---",
        **kwargs,
    ):

        super(QuerySetSelectMultipleField, self).__init__(
            label, validators, queryset, label_attr, allow_blank, blank_text, **kwargs
        )

    def process_formdata(self, valuelist):

        if valuelist:
            if valuelist[0] == "__None":
                self.data = None
            else:
                if not self.queryset:
                    self.data = None
                    return

                self.queryset.rewind()
                self.data = list(self.queryset(pk__in=valuelist))
                if not len(self.data):
                    self.data = None

        # If no value passed, empty the list
        else:
            self.data = None

    def _is_selected(self, item):
        return item in self.data if self.data else False


class ModelSelectField(QuerySetSelectField):
    """
    Like a QuerySetSelectField, except takes a model class instead of a
    queryset and lists everything in it.
    """

    def __init__(self, label="", validators=None, model=None, **kwargs):
        queryset = kwargs.pop("queryset", model.objects)
        super(ModelSelectField, self).__init__(
            label, validators, queryset=queryset, **kwargs
        )


class ModelSelectMultipleField(QuerySetSelectMultipleField):
    """
    Allows multiple select
    """

    def __init__(self, label="", validators=None, model=None, **kwargs):
        queryset = kwargs.pop("queryset", model.objects)
        super(ModelSelectMultipleField, self).__init__(
            label, validators, queryset=queryset, **kwargs
        )


class JSONField(TextAreaField):
    def _value(self):
        if self.raw_data:
            return self.raw_data[0]
        else:
            return self.data and unicode(json.dumps(self.data)) or ""

    def process_formdata(self, value):
        if value:
            try:
                self.data = json.loads(value[0])
            except ValueError:
                raise ValueError(self.gettext("Invalid JSON data."))


class DictField(JSONField):
    def process_formdata(self, value):
        super(DictField, self).process_formdata(value)
        if value and not isinstance(self.data, dict):
            raise ValueError(self.gettext("Not a valid dictionary."))


class NoneStringField(StringField):
    """
    Custom StringField that counts "" as None
    """

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = valuelist[0]
        if self.data == "":
            self.data = None


class BinaryField(TextAreaField):
    """
    Custom TextAreaField that converts its value with binary_type.
    """

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = bytes(valuelist[0], "utf-8")



import datetime
import os
import sys

import flask

sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), "../../")))

from flask_mongoengine import MongoEngine
from flask_debugtoolbar import DebugToolbarExtension

app = flask.Flask(__name__)
app.config.from_object(__name__)
app.config["MONGODB_SETTINGS"] = {"DB": "testing"}
app.config["TESTING"] = True
app.config["SECRET_KEY"] = "flask+mongoengine=<3"
app.debug = True
app.config["DEBUG_TB_PANELS"] = (
    "flask_debugtoolbar.panels.versions.VersionDebugPanel",
    "flask_debugtoolbar.panels.timer.TimerDebugPanel",
    "flask_debugtoolbar.panels.headers.HeaderDebugPanel",
    "flask_debugtoolbar.panels.request_vars.RequestVarsDebugPanel",
    "flask_debugtoolbar.panels.template.TemplateDebugPanel",
    "flask_debugtoolbar.panels.logger.LoggingPanel",
    "flask_mongoengine.panels.MongoDebugPanel",
)

app.config["DEBUG_TB_INTERCEPT_REDIRECTS"] = False

db = MongoEngine()
db.init_app(app)

DebugToolbarExtension(app)


class Todo(db.Document):
    title = db.StringField(max_length=60)
    text = db.StringField()
    done = db.BooleanField(default=False)
    pub_date = db.DateTimeField(default=datetime.datetime.now)


@app.route("/")
def index():
    # As a list to test debug toolbar
    Todo.objects().delete()  # Removes
    Todo(title="Simple todo A", text="12345678910").save()  # Insert
    Todo(title="Simple todo B", text="12345678910").save()  # Insert
    Todo.objects(title__contains="B").update(set__text="Hello world")  # Update
    todos = Todo.objects.all()
    return flask.render_template("index.html", todos=todos)


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=4000)



"""
Tools for generating forms based on mongoengine Document schemas.
"""
import decimal
import sys
from collections import OrderedDict

from bson import ObjectId
from mongoengine import ReferenceField
from wtforms import fields as f, validators

from flask_mongoengine.wtf.fields import (
    BinaryField,
    DictField,
    ModelSelectField,
    ModelSelectMultipleField,
    NoneStringField,
)
from flask_mongoengine.wtf.models import ModelForm

__all__ = (
    "model_fields",
    "model_form",
)


if sys.version_info >= (3, 0):
    unicode = str


def converts(*args):
    def _inner(func):
        func._converter_for = frozenset(args)
        return func

    return _inner


class ModelConverter(object):
    def __init__(self, converters=None):
        if not converters:
            converters = {}

        for name in dir(self):
            obj = getattr(self, name)
            if hasattr(obj, "_converter_for"):
                for classname in obj._converter_for:
                    converters[classname] = obj

        self.converters = converters

    def convert(self, model, field, field_args):
        kwargs = {
            "label": getattr(field, "verbose_name", field.name),
            "description": getattr(field, "help_text", None) or "",
            "validators": getattr(field, "validators", None) or [],
            "filters": getattr(field, "filters", None) or [],
            "default": field.default,
        }
        if field_args:
            kwargs.update(field_args)

        if kwargs["validators"]:
            # Create a copy of the list since we will be modifying it.
            kwargs["validators"] = list(kwargs["validators"])

        if field.required:
            kwargs["validators"].append(validators.InputRequired())
        else:
            kwargs["validators"].append(validators.Optional())

        ftype = type(field).__name__

        if field.choices:
            kwargs["choices"] = field.choices

            if ftype in self.converters:
                kwargs["coerce"] = self.coerce(ftype)
            multiple_field = kwargs.pop("multiple", False)
            radio_field = kwargs.pop("radio", False)
            if multiple_field:
                return f.SelectMultipleField(**kwargs)
            if radio_field:
                return f.RadioField(**kwargs)
            return f.SelectField(**kwargs)

        ftype = type(field).__name__

        if hasattr(field, "to_form_field"):
            return field.to_form_field(model, kwargs)

        if hasattr(field, "field") and type(field.field) == ReferenceField:
            kwargs["label_modifier"] = getattr(
                model, field.name + "_label_modifier", None
            )

        if ftype in self.converters:
            return self.converters[ftype](model, field, kwargs)

    @classmethod
    def _string_common(cls, model, field, kwargs):
        if field.max_length or field.min_length:
            kwargs["validators"].append(
                validators.Length(
                    max=field.max_length or -1, min=field.min_length or -1
                )
            )

    @classmethod
    def _number_common(cls, model, field, kwargs):
        if field.max_value or field.min_value:
            kwargs["validators"].append(
                validators.NumberRange(max=field.max_value, min=field.min_value)
            )

    @converts("StringField")
    def conv_String(self, model, field, kwargs):
        if field.regex:
            kwargs["validators"].append(validators.Regexp(regex=field.regex))
        self._string_common(model, field, kwargs)
        password_field = kwargs.pop("password", False)
        textarea_field = kwargs.pop("textarea", False) or not field.max_length
        if password_field:
            return f.PasswordField(**kwargs)
        if textarea_field:
            return f.TextAreaField(**kwargs)
        return f.StringField(**kwargs)

    @converts("URLField")
    def conv_URL(self, model, field, kwargs):
        kwargs["validators"].append(validators.URL())
        self._string_common(model, field, kwargs)
        return NoneStringField(**kwargs)

    @converts("EmailField")
    def conv_Email(self, model, field, kwargs):
        kwargs["validators"].append(validators.Email())
        self._string_common(model, field, kwargs)
        return NoneStringField(**kwargs)

    @converts("IntField")
    def conv_Int(self, model, field, kwargs):
        self._number_common(model, field, kwargs)
        return f.IntegerField(**kwargs)

    @converts("FloatField")
    def conv_Float(self, model, field, kwargs):
        self._number_common(model, field, kwargs)
        return f.FloatField(**kwargs)

    @converts("DecimalField")
    def conv_Decimal(self, model, field, kwargs):
        self._number_common(model, field, kwargs)
        kwargs["places"] = getattr(field, "precision", None)
        return f.DecimalField(**kwargs)

    @converts("BooleanField")
    def conv_Boolean(self, model, field, kwargs):
        return f.BooleanField(**kwargs)

    @converts("DateTimeField")
    def conv_DateTime(self, model, field, kwargs):
        return f.DateTimeField(**kwargs)

    @converts("DateField")
    def conv_Date(self, model, field, kwargs):
        return f.DateField(**kwargs)

    @converts("BinaryField")
    def conv_Binary(self, model, field, kwargs):
        # TODO: may be set file field that will save file`s data to MongoDB
        if field.max_bytes:
            kwargs["validators"].append(validators.Length(max=field.max_bytes))
        return BinaryField(**kwargs)

    @converts("DictField")
    def conv_Dict(self, model, field, kwargs):
        return DictField(**kwargs)

    @converts("ListField")
    def conv_List(self, model, field, kwargs):
        if isinstance(field.field, ReferenceField):
            return ModelSelectMultipleField(model=field.field.document_type, **kwargs)
        if field.field.choices:
            kwargs["multiple"] = True
            return self.convert(model, field.field, kwargs)
        field_args = kwargs.pop("field_args", {})
        unbound_field = self.convert(model, field.field, field_args)
        unacceptable = {
            "validators": [],
            "filters": [],
            "min_entries": kwargs.get("min_entries", 0),
        }
        kwargs.update(unacceptable)
        return f.FieldList(unbound_field, **kwargs)

    @converts("SortedListField")
    def conv_SortedList(self, model, field, kwargs):
        # TODO: sort functionality, may be need sortable widget
        return self.conv_List(model, field, kwargs)

    @converts("GeoLocationField")
    def conv_GeoLocation(self, model, field, kwargs):
        # TODO: create geo field and widget (also GoogleMaps)
        return

    @converts("ObjectIdField")
    def conv_ObjectId(self, model, field, kwargs):
        return

    @converts("EmbeddedDocumentField")
    def conv_EmbeddedDocument(self, model, field, kwargs):
        kwargs = {
            "validators": [],
            "filters": [],
            "default": field.default or field.document_type_obj,
        }
        form_class = model_form(field.document_type_obj, field_args={})
        return f.FormField(form_class, **kwargs)

    @converts("ReferenceField")
    def conv_Reference(self, model, field, kwargs):
        return ModelSelectField(model=field.document_type, **kwargs)

    @converts("GenericReferenceField")
    def conv_GenericReference(self, model, field, kwargs):
        return

    def coerce(self, field_type):
        coercions = {
            "IntField": int,
            "BooleanField": bool,
            "FloatField": float,
            "DecimalField": decimal.Decimal,
            "ObjectIdField": ObjectId,
        }
        return coercions.get(field_type, unicode)


def model_fields(model, only=None, exclude=None, field_args=None, converter=None):
    """
    Generate a dictionary of fields for a given database model.

    See `model_form` docstring for description of parameters.
    """
    from mongoengine.base import BaseDocument, DocumentMetaclass

    if not isinstance(model, (BaseDocument, DocumentMetaclass)):
        raise TypeError("model must be a mongoengine Document schema")

    converter = converter or ModelConverter()
    field_args = field_args or {}

    names = ((k, v.creation_counter) for k, v in model._fields.items())
    field_names = [n[0] for n in sorted(names, key=lambda n: n[1])]

    if only:
        field_names = [x for x in only if x in set(field_names)]
    elif exclude:
        field_names = [x for x in field_names if x not in set(exclude)]

    field_dict = OrderedDict()
    for name in field_names:
        model_field = model._fields[name]
        field = converter.convert(model, model_field, field_args.get(name))
        if field is not None:
            field_dict[name] = field

    return field_dict


def model_form(
    model,
    base_class=ModelForm,
    only=None,
    exclude=None,
    field_args=None,
    converter=None,
):
    """
    Create a wtforms Form for a given mongoengine Document schema::

        from flask_mongoengine.wtf import model_form
        from myproject.myapp.schemas import Article
        ArticleForm = model_form(Article)

    :param model:
        A mongoengine Document schema class
    :param base_class:
        Base form class to extend from. Must be a ``wtforms.Form`` subclass.
    :param only:
        An optional iterable with the property names that should be included in
        the form. Only these properties will have fields.
    :param exclude:
        An optional iterable with the property names that should be excluded
        from the form. All other properties will have fields.
    :param field_args:
        An optional dictionary of field names mapping to keyword arguments used
        to construct each field object.
    :param converter:
        A converter to generate the fields based on the model properties. If
        not set, ``ModelConverter`` is used.
    """
    field_dict = model_fields(model, only, exclude, field_args, converter)
    field_dict["model_class"] = model
    return type(model.__name__ + "Form", (base_class,), field_dict)



from flask_wtf import FlaskForm
from flask_wtf.form import _Auto


class ModelForm(FlaskForm):
    """A WTForms mongoengine model form"""

    def __init__(self, formdata=_Auto, **kwargs):
        self.instance = kwargs.pop("instance", None) or kwargs.get("obj")
        if self.instance and not formdata:
            kwargs["obj"] = self.instance
        self.formdata = formdata
        super(ModelForm, self).__init__(formdata, **kwargs)

    def save(self, commit=True, **kwargs):
        if self.instance:
            self.populate_obj(self.instance)
        else:
            self.instance = self.model_class(**self.data)

        if commit:
            self.instance.save(**kwargs)
        return self.instance



import pytest
from flask import session

from flask_mongoengine import MongoEngineSessionInterface


@pytest.fixture(autouse=True)
def setup_endpoints(app, db):

    app.session_interface = MongoEngineSessionInterface(db)

    @app.route("/")
    def index():
        session["a"] = "hello session"
        return session["a"]

    @app.route("/check-session")
    def check_session():
        return "session: %s" % session["a"]

    @app.route("/check-session-database")
    def check_session_database():
        sessions = app.session_interface.cls.objects.count()
        return "sessions: %s" % sessions


def test_setting_session(app):
    client = app.test_client()

    response = client.get("/")
    assert response.status_code == 200
    assert response.data.decode("utf-8") == "hello session"

    response = client.get("/check-session")
    assert response.status_code == 200
    assert response.data.decode("utf-8") == "session: hello session"

    response = client.get("/check-session-database")
    assert response.status_code == 200
    assert response.data.decode("utf-8") == "sessions: 1"



from datetime import datetime

import mongoengine
import pytest
from flask import Flask

from flask_mongoengine import MongoEngine


@pytest.fixture()
def app():
    app = Flask(__name__)
    app.config["TESTING"] = True
    app.config["WTF_CSRF_ENABLED"] = False

    with app.app_context():
        yield app

    mongoengine.connection.disconnect_all()


@pytest.fixture()
def db(app):
    app.config["MONGODB_HOST"] = "mongodb://localhost:27017/flask_mongoengine_test_db"
    test_db = MongoEngine(app)
    db_name = test_db.connection.get_database("flask_mongoengine_test_db").name

    if not db_name.endswith("_test_db"):
        raise RuntimeError(
            f"DATABASE_URL must point to testing db, not to master db ({db_name})"
        )

    # Clear database before tests, for cases when some test failed before.
    test_db.connection.drop_database(db_name)

    yield test_db

    # Clear database after tests, for graceful exit.
    test_db.connection.drop_database(db_name)


@pytest.fixture()
def todo(db):
    class Todo(db.Document):
        title = mongoengine.StringField(max_length=60)
        text = mongoengine.StringField()
        done = mongoengine.BooleanField(default=False)
        pub_date = mongoengine.DateTimeField(default=datetime.utcnow)
        comments = mongoengine.ListField(mongoengine.StringField())
        comment_count = mongoengine.IntField()

    return Todo



import os
import sys

import flask

sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), "../../")))

from flask_debugtoolbar import DebugToolbarExtension

app = flask.Flask(__name__)
app.config.from_object(__name__)
app.config["MONGODB_SETTINGS"] = {"DB": "testing"}
app.config["TESTING"] = True
app.config["SECRET_KEY"] = "flask+mongoengine=<3"
app.debug = True
app.config["DEBUG_TB_PANELS"] = (
    "flask_debugtoolbar.panels.versions.VersionDebugPanel",
    "flask_debugtoolbar.panels.timer.TimerDebugPanel",
    "flask_debugtoolbar.panels.headers.HeaderDebugPanel",
    "flask_debugtoolbar.panels.request_vars.RequestVarsDebugPanel",
    "flask_debugtoolbar.panels.template.TemplateDebugPanel",
    "flask_debugtoolbar.panels.logger.LoggingPanel",
    "flask_mongoengine.panels.MongoDebugPanel",
)
app.config["DEBUG_TB_INTERCEPT_REDIRECTS"] = False

from models import db

db.init_app(app)

DebugToolbarExtension(app)

from views import index, pagination

app.add_url_rule("/", view_func=index)
app.add_url_rule("/pagination", view_func=pagination)

if __name__ == "__main__":
    app.run(host="0.0.0.0", port=4000)



from mongoengine.base import BaseField

__all__ = "WtfBaseField"


class WtfBaseField(BaseField):
    """
    Extension wrapper class for mongoengine BaseField.

    This enables flask-mongoengine  wtf to extend the
    number of field parameters, and settings on behalf
    of document model form generator for WTForm.

    @param validators:  wtf model form field validators.
    @param filters:     wtf model form field filters.
    """

    def __init__(self, validators=None, filters=None, **kwargs):

        self.validators = self._ensure_callable_or_list(validators, "validators")
        self.filters = self._ensure_callable_or_list(filters, "filters")

        BaseField.__init__(self, **kwargs)

    def _ensure_callable_or_list(self, field, msg_flag):
        """
        Ensure the value submitted via field is either
        a callable object to convert to list or it is
        in fact a valid list value.

        """
        if field is not None:
            if callable(field):
                field = [field]
            else:
                msg = "Argument '%s' must be a list value" % msg_flag
                if not isinstance(field, list):
                    raise TypeError(msg)

        return field



import datetime
import sys
import uuid

from bson.tz_util import utc
from flask.sessions import SessionInterface, SessionMixin
from werkzeug.datastructures import CallbackDict

__all__ = ("MongoEngineSession", "MongoEngineSessionInterface")

if sys.version_info >= (3, 0):
    basestring = str


class MongoEngineSession(CallbackDict, SessionMixin):
    def __init__(self, initial=None, sid=None):
        def on_update(self):
            self.modified = True

        CallbackDict.__init__(self, initial, on_update)
        self.sid = sid
        self.modified = False


class MongoEngineSessionInterface(SessionInterface):
    """SessionInterface for mongoengine"""

    def __init__(self, db, collection="session"):
        """
        The MongoSessionInterface

        :param db: The app's db eg: MongoEngine()
        :param collection: The session collection name defaults to "session"
        """

        if not isinstance(collection, basestring):
            raise ValueError("collection argument should be string or unicode")

        class DBSession(db.Document):
            sid = db.StringField(primary_key=True)
            data = db.DictField()
            expiration = db.DateTimeField()
            meta = {
                "allow_inheritance": False,
                "collection": collection,
                "indexes": [
                    {
                        "fields": ["expiration"],
                        "expireAfterSeconds": 60 * 60 * 24 * 7 * 31,
                    }
                ],
            }

        self.cls = DBSession

    def get_expiration_time(self, app, session):
        if session.permanent:
            return app.permanent_session_lifetime
        if "SESSION_TTL" in app.config:
            return datetime.timedelta(**app.config["SESSION_TTL"])
        return datetime.timedelta(days=1)

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if sid:
            stored_session = self.cls.objects(sid=sid).first()

            if stored_session:
                expiration = stored_session.expiration

                if not expiration.tzinfo:
                    expiration = expiration.replace(tzinfo=utc)

                if expiration > datetime.datetime.utcnow().replace(tzinfo=utc):
                    return MongoEngineSession(
                        initial=stored_session.data, sid=stored_session.sid
                    )

        return MongoEngineSession(sid=str(uuid.uuid4()))

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        httponly = self.get_cookie_httponly(app)

        if not session:
            if session.modified:
                response.delete_cookie(app.session_cookie_name, domain=domain)
            return

        expiration = datetime.datetime.utcnow().replace(
            tzinfo=utc
        ) + self.get_expiration_time(app, session)

        if session.modified:
            self.cls(sid=session.sid, data=session, expiration=expiration).save()

        response.set_cookie(
            app.session_cookie_name,
            session.sid,
            expires=expiration,
            httponly=httponly,
            domain=domain,
        )



import math

from flask import abort
from mongoengine.queryset import QuerySet

__all__ = ("Pagination", "ListFieldPagination")


class Pagination(object):
    def __init__(self, iterable, page, per_page):

        if page < 1:
            abort(404)

        self.iterable = iterable
        self.page = page
        self.per_page = per_page

        if isinstance(iterable, QuerySet):
            self.total = iterable.count()
        else:
            self.total = len(iterable)

        start_index = (page - 1) * per_page
        end_index = page * per_page

        self.items = iterable[start_index:end_index]
        if isinstance(self.items, QuerySet):
            self.items = self.items.select_related()
        if not self.items and page != 1:
            abort(404)

    @property
    def pages(self):
        """The total number of pages"""
        return int(math.ceil(self.total / float(self.per_page)))

    def prev(self, error_out=False):
        """Returns a :class:`Pagination` object for the previous page."""
        assert self.iterable is not None, (
            "an object is required " "for this method to work"
        )
        iterable = self.iterable
        if isinstance(iterable, QuerySet):
            iterable._skip = None
            iterable._limit = None
        return self.__class__(iterable, self.page - 1, self.per_page)

    @property
    def prev_num(self):
        """Number of the previous page."""
        return self.page - 1

    @property
    def has_prev(self):
        """True if a previous page exists"""
        return self.page > 1

    def next(self, error_out=False):
        """Returns a :class:`Pagination` object for the next page."""
        assert self.iterable is not None, (
            "an object is required " "for this method to work"
        )
        iterable = self.iterable
        if isinstance(iterable, QuerySet):
            iterable._skip = None
            iterable._limit = None
        return self.__class__(iterable, self.page + 1, self.per_page)

    @property
    def has_next(self):
        """True if a next page exists."""
        return self.page < self.pages

    @property
    def next_num(self):
        """Number of the next page"""
        return self.page + 1

    def iter_pages(self, left_edge=2, left_current=2, right_current=5, right_edge=2):
        """Iterates over the page numbers in the pagination.  The four
        parameters control the thresholds how many numbers should be produced
        from the sides.  Skipped page numbers are represented as `None`.
        This is how you could render such a pagination in the templates:

        .. sourcecode:: html+jinja

            {% macro render_pagination(pagination, endpoint) %}
              <div class=pagination>
              {%- for page in pagination.iter_pages() %}
                {% if page %}
                  {% if page != pagination.page %}
                    <a href="{{ url_for(endpoint, page=page) }}">{{ page }}</a>
                  {% else %}
                    <strong>{{ page }}</strong>
                  {% endif %}
                {% else %}
                  <span class=ellipsis>…</span>
                {% endif %}
              {%- endfor %}
              </div>
            {% endmacro %}
        """
        last = 0
        for num in range(1, self.pages + 1):
            if (
                num <= left_edge
                or num > self.pages - right_edge
                or (
                    num >= self.page - left_current and num <= self.page + right_current
                )
            ):
                if last + 1 != num:
                    yield None
                yield num
                last = num
        if last != self.pages:
            yield None


class ListFieldPagination(Pagination):
    def __init__(self, queryset, doc_id, field_name, page, per_page, total=None):
        """Allows an array within a document to be paginated.

        Queryset must contain the document which has the array we're
        paginating, and doc_id should be it's _id.
        Field name is the name of the array we're paginating.
        Page and per_page work just like in Pagination.
        Total is an argument because it can be computed more efficiently
        elsewhere, but we still use array.length as a fallback.
        """
        if page < 1:
            abort(404)

        self.page = page
        self.per_page = per_page

        self.queryset = queryset
        self.doc_id = doc_id
        self.field_name = field_name

        start_index = (page - 1) * per_page

        field_attrs = {field_name: {"$slice": [start_index, per_page]}}

        qs = queryset(pk=doc_id)
        self.items = getattr(qs.fields(**field_attrs).first(), field_name)
        self.total = total or len(
            getattr(qs.fields(**{field_name: 1}).first(), field_name)
        )

        if not self.items and page != 1:
            abort(404)

    def prev(self, error_out=False):
        """Returns a :class:`Pagination` object for the previous page."""
        assert self.items is not None, (
            "a query object is required " "for this method to work"
        )
        return self.__class__(
            self.queryset,
            self.doc_id,
            self.field_name,
            self.page - 1,
            self.per_page,
            self.total,
        )

    def next(self, error_out=False):
        """Returns a :class:`Pagination` object for the next page."""
        assert self.items is not None, (
            "a query object is required " "for this method to work"
        )
        return self.__class__(
            self.queryset,
            self.doc_id,
            self.field_name,
            self.page + 1,
            self.per_page,
            self.total,
        )



from bson import json_util
from flask.json import JSONEncoder
from mongoengine.base import BaseDocument
from mongoengine.queryset import QuerySet


def _make_encoder(superclass):
    class MongoEngineJSONEncoder(superclass):
        """
        A JSONEncoder which provides serialization of MongoEngine
        documents and queryset objects.
        """

        def default(self, obj):
            if isinstance(obj, BaseDocument):
                return json_util._json_convert(obj.to_mongo())
            elif isinstance(obj, QuerySet):
                return json_util._json_convert(obj.as_pymongo())
            return superclass.default(self, obj)

    return MongoEngineJSONEncoder


MongoEngineJSONEncoder = _make_encoder(JSONEncoder)


def override_json_encoder(app):
    """
    A function to dynamically create a new MongoEngineJSONEncoder class
    based upon a custom base class.
    This function allows us to combine MongoEngine serialization with
    any changes to Flask's JSONEncoder which a user may have made
    prior to calling init_app.

    NOTE: This does not cover situations where users override
    an instance's json_encoder after calling init_app.
    """
    app.json_encoder = _make_encoder(app.json_encoder)



from flask import current_app
from flask_debugtoolbar.panels import DebugPanel
from jinja2 import ChoiceLoader, PackageLoader

from flask_mongoengine import operation_tracker

package_loader = PackageLoader("flask_mongoengine", "templates")


def _maybe_patch_jinja_loader(jinja_env):
    """Patch the jinja_env loader to include flaskext.mongoengine
    templates folder if necessary.
    """
    if not isinstance(jinja_env.loader, ChoiceLoader):
        jinja_env.loader = ChoiceLoader([jinja_env.loader, package_loader])
    elif package_loader not in jinja_env.loader.loaders:
        jinja_env.loader.loaders.append(package_loader)


class MongoDebugPanel(DebugPanel):
    """Panel that shows information about MongoDB operations (including stack)

    Adapted from https://github.com/hmarr/django-debug-toolbar-mongo
    """

    name = "MongoDB"
    has_content = True

    def __init__(self, *args, **kwargs):
        super(MongoDebugPanel, self).__init__(*args, **kwargs)
        _maybe_patch_jinja_loader(self.jinja_env)
        operation_tracker.install_tracker()

    def process_request(self, request):
        operation_tracker.reset()

    def nav_title(self):
        return "MongoDB"

    def nav_subtitle(self):
        attrs = ["queries", "inserts", "updates", "removes"]
        ops = sum(
            sum((1 for o in getattr(operation_tracker, a) if not o["internal"]))
            for a in attrs
        )
        total_time = sum(
            sum(o["time"] for o in getattr(operation_tracker, a)) for a in attrs
        )
        return "{0} operations in {1:.2f}ms".format(ops, total_time)

    def title(self):
        return "MongoDB Operations"

    def url(self):
        return ""

    def content(self):
        context = self.context.copy()
        context["queries"] = operation_tracker.queries
        context["inserts"] = operation_tracker.inserts
        context["updates"] = operation_tracker.updates
        context["removes"] = operation_tracker.removes
        context["slow_query_limit"] = current_app.config.get(
            "MONGO_DEBUG_PANEL_SLOW_QUERY_LIMIT", 100
        )
        return self.render("panels/mongo-panel.html", context)



import datetime

from flask_mongoengine import MongoEngine

db = MongoEngine()


class Todo(db.Document):
    title = db.StringField(max_length=60)
    text = db.StringField()
    done = db.BooleanField(default=False)
    pub_date = db.DateTimeField(default=datetime.datetime.now)



"""Tests for base MongoEngine class."""
from flask_mongoengine import MongoEngine
import pytest


def test_mongoengine_class__should_raise_type_error__if_config_not_dict():
    """MongoEngine will handle None values, but will pass anything else as app."""
    input_value = "Not dict type"
    with pytest.raises(TypeError) as error:
        MongoEngine(input_value)
    assert str(error.value) == "Invalid Flask application instance"


@pytest.mark.parametrize("input_value", [None, "Not dict type"])
def test_init_app__should_raise_type_error__if_config_not_dict(input_value):
    db = MongoEngine()
    with pytest.raises(TypeError) as error:
        db.init_app(input_value)
    assert str(error.value) == "Invalid Flask application instance"



import flask
import pytest
from bson import ObjectId


@pytest.fixture(autouse=True)
def setup_endpoints(app, todo):
    Todo = todo

    @app.route("/")
    def index():
        return flask.jsonify(result=Todo.objects())

    @app.route("/add", methods=["POST"])
    def add():
        form = flask.request.form
        todo = Todo(title=form["title"], text=form["text"])
        todo.save()
        return flask.jsonify(result=todo)

    @app.route("/show/<id>/")
    def show(id):
        return flask.jsonify(result=Todo.objects.get_or_404(id=id))


def test_with_id(app, todo):
    Todo = todo
    client = app.test_client()
    response = client.get("/show/%s/" % ObjectId())
    assert response.status_code == 404

    response = client.post("/add", data={"title": "First Item", "text": "The text"})
    assert response.status_code == 200

    response = client.get("/show/%s/" % Todo.objects.first().id)
    assert response.status_code == 200

    result = flask.json.loads(response.data).get("result")
    assert ("title", "First Item") in result.items()


def test_basic_insert(app):
    client = app.test_client()
    client.post("/add", data={"title": "First Item", "text": "The text"})
    client.post("/add", data={"title": "2nd Item", "text": "The text"})
    rv = client.get("/")
    result = flask.json.loads(rv.data).get("result")

    assert len(result) == 2



import pytest
from werkzeug.exceptions import NotFound

from flask_mongoengine import ListFieldPagination, Pagination


def test_queryset_paginator(app, todo):
    Todo = todo
    for i in range(42):
        Todo(title="post: %s" % i).save()

    with pytest.raises(NotFound):
        Pagination(iterable=Todo.objects, page=0, per_page=10)

    with pytest.raises(NotFound):
        Pagination(iterable=Todo.objects, page=6, per_page=10)

    paginator = Pagination(Todo.objects, 1, 10)
    _test_paginator(paginator)


def test_paginate_plain_list():
    with pytest.raises(NotFound):
        Pagination(iterable=range(1, 42), page=0, per_page=10)

    with pytest.raises(NotFound):
        Pagination(iterable=range(1, 42), page=6, per_page=10)

    paginator = Pagination(range(1, 42), 1, 10)
    _test_paginator(paginator)


def test_list_field_pagination(app, todo):
    Todo = todo

    comments = ["comment: %s" % i for i in range(42)]
    todo = Todo(
        title="todo has comments",
        comments=comments,
        comment_count=len(comments),
    ).save()

    # Check without providing a total
    paginator = ListFieldPagination(Todo.objects, todo.id, "comments", 1, 10)
    _test_paginator(paginator)

    # Check with providing a total (saves a query)
    paginator = ListFieldPagination(
        Todo.objects, todo.id, "comments", 1, 10, todo.comment_count
    )
    _test_paginator(paginator)

    paginator = todo.paginate_field("comments", 1, 10)
    _test_paginator(paginator)


def _test_paginator(paginator):
    assert 5 == paginator.pages
    assert [1, 2, 3, 4, 5] == list(paginator.iter_pages())

    for i in [1, 2, 3, 4, 5]:

        if i == 1:
            assert not paginator.has_prev
            with pytest.raises(NotFound):
                paginator.prev()
        else:
            assert paginator.has_prev

        if i == 5:
            assert not paginator.has_next
            with pytest.raises(NotFound):
                paginator.next()
        else:
            assert paginator.has_next

        if i == 3:
            assert [None, 2, 3, 4, None] == list(paginator.iter_pages(0, 1, 1, 0))

        assert i == paginator.page
        assert i - 1 == paginator.prev_num
        assert i + 1 == paginator.next_num

        # Paginate to the next page
        if i < 5:
            paginator = paginator.next()



import datetime
import re

import bson
import flask
import pytest
import wtforms
from mongoengine import NotUniqueError, queryset_manager
from werkzeug.datastructures import MultiDict

from flask_mongoengine.wtf import model_form


def test_binaryfield(app, db):

    with app.test_request_context("/"):

        class Binary(db.Document):
            binary = db.BinaryField()

        BinaryForm = model_form(Binary)
        form = BinaryForm(MultiDict({"binary": "1"}))
        assert form.validate()
        form.save()


def test_choices_coerce(app, db):

    with app.test_request_context("/"):

        CHOICES = ((1, "blue"), (2, "red"))

        class MyChoices(db.Document):
            pill = db.IntField(choices=CHOICES)

        MyChoicesForm = model_form(MyChoices)
        form = MyChoicesForm(MultiDict({"pill": "1"}))
        assert form.validate()
        form.save()
        assert MyChoices.objects.first().pill == 1


def test_list_choices_coerce(app, db):

    with app.test_request_context("/"):

        CHOICES = ((1, "blue"), (2, "red"))

        class MyChoices(db.Document):
            pill = db.ListField(db.IntField(choices=CHOICES))

        MyChoicesForm = model_form(MyChoices)
        form = MyChoicesForm(MultiDict({"pill": "1"}))
        assert form.validate()
        form.save()
        assert MyChoices.objects.first().pill[0] == 1


def test_emailfield(app, db):

    with app.test_request_context("/"):

        class Email(db.Document):
            email = db.EmailField(required=False)

        EmailForm = model_form(Email)
        form = EmailForm(instance=Email())
        assert "None" not in "%s" % form.email
        assert form.validate()

        form = EmailForm(MultiDict({"email": ""}))
        assert "None" not in "%s" % form.email
        assert form.validate()

        # Ensure required works

        class Email(db.Document):
            email = db.EmailField(required=True)

        EmailForm = model_form(Email)
        form = EmailForm(MultiDict({"email": ""}))
        assert "None" not in "%s" % form.email
        assert not form.validate()


def test_model_form(app, db):
    with app.test_request_context("/"):

        class BlogPost(db.Document):
            meta = {"allow_inheritance": True}
            title = db.StringField(required=True, max_length=200)
            posted = db.DateTimeField(default=datetime.datetime.now)
            tags = db.ListField(db.StringField())

        class TextPost(BlogPost):
            email = db.EmailField(required=False)
            lead_paragraph = db.StringField(max_length=200)
            content = db.StringField(required=True)

        class LinkPost(BlogPost):
            url = db.StringField(required=True, max_length=200)
            interest = db.DecimalField(required=True)

        # Create a text-based post
        TextPostForm = model_form(
            TextPost, field_args={"lead_paragraph": {"textarea": True}}
        )

        form = TextPostForm(
            MultiDict(
                {"title": "Using MongoEngine", "tags": ["mongodb", "mongoengine"]}
            )
        )

        assert not form.validate()

        form = TextPostForm(
            MultiDict(
                {
                    "title": "Using MongoEngine",
                    "content": "See the tutorial",
                    "tags": ["mongodb", "mongoengine"],
                }
            )
        )

        assert form.validate()
        form.save()

        assert form.title.type == "StringField"
        assert form.content.type == "TextAreaField"
        assert form.lead_paragraph.type == "TextAreaField"

        assert BlogPost.objects.first().title == "Using MongoEngine"
        assert BlogPost.objects.count() == 1

        form = TextPostForm(
            MultiDict(
                {
                    "title": "Using Flask-MongoEngine",
                    "content": "See the tutorial",
                    "tags": ["flask", "mongodb", "mongoengine"],
                }
            )
        )

        assert form.validate()
        form.save()
        assert BlogPost.objects.count() == 2

        post = BlogPost.objects(title="Using Flask-MongoEngine").get()

        form = TextPostForm(
            MultiDict(
                {
                    "title": "Using Flask-MongoEngine",
                    "content": "See the tutorial",
                    "tags-0": "flask",
                    "tags-1": "mongodb",
                    "tags-2": "mongoengine",
                    "tags-3": "flask-mongoengine",
                }
            ),
            instance=post,
        )
        assert form.validate()
        form.save()
        post = post.reload()

        assert post.tags == ["flask", "mongodb", "mongoengine", "flask-mongoengine"]

        # Create a link post
        LinkPostForm = model_form(LinkPost)

        form = LinkPostForm(
            MultiDict(
                {
                    "title": "Using Flask-MongoEngine",
                    "url": "http://flask-mongoengine.org",
                    "interest": "0",
                }
            )
        )
        form.validate()
        assert form.validate()


def test_model_form_only(app, db):
    with app.test_request_context("/"):

        class BlogPost(db.Document):
            title = db.StringField(required=True, max_length=200)
            posted = db.DateTimeField(default=datetime.datetime.now)
            tags = db.ListField(db.StringField())

        BlogPost.drop_collection()

        BlogPostForm = model_form(BlogPost, only=["tags"])
        form = BlogPostForm()
        assert hasattr(form, "tags")
        assert not hasattr(form, "posted")

        BlogPostForm = model_form(BlogPost, exclude=["posted"])
        form = BlogPostForm()
        assert hasattr(form, "tags")
        assert not hasattr(form, "posted")


def test_model_form_with_custom_query_set(app, db):
    with app.test_request_context("/"):

        class Dog(db.Document):
            breed = db.StringField()

            @queryset_manager
            def large_objects(cls, queryset):
                return queryset(breed__in=["german sheppard", "wolfhound"])

        class DogOwner(db.Document):
            dog = db.ReferenceField(Dog)

        big_dogs = [Dog(breed="german sheppard"), Dog(breed="wolfhound")]
        dogs = [Dog(breed="poodle")] + big_dogs
        for dog in dogs:
            dog.save()

        BigDogForm = model_form(
            DogOwner, field_args={"dog": {"queryset": Dog.large_objects}}
        )

        form = BigDogForm(dog=big_dogs[0])
        assert form.validate()
        assert big_dogs == [d[1] for d in form.dog.iter_choices()]


def test_modelselectfield(app, db):
    with app.test_request_context("/"):

        class Dog(db.Document):
            name = db.StringField()

        class DogOwner(db.Document):
            dog = db.ReferenceField(Dog)

        DogOwnerForm = model_form(DogOwner, field_args={"dog": {"allow_blank": True}})

        dog = Dog(name="fido")
        dog.save()

        form = DogOwnerForm(dog=dog)
        assert form.validate()

        assert isinstance(form.dog.widget, wtforms.widgets.Select)
        assert not form.dog.widget.multiple

        # Validate the options - should contain a dog (selected) and a
        # blank option there should be an extra blank option.
        choices = list(form.dog)
        assert len(choices) == 2
        assert not choices[0].checked
        assert choices[0].data == "__None"
        assert choices[1].checked
        assert choices[1].data == dog.pk

        # Validate selecting one item
        form = DogOwnerForm(MultiDict({"dog": dog.id}))
        assert form.dog.data == dog

        # Validate selecting no item
        form = DogOwnerForm(MultiDict({"dog": "__None"}), dog=dog)
        assert form.dog.data is None


def test_modelselectfield_multiple(app, db):
    with app.test_request_context("/"):

        class Dog(db.Document):
            name = db.StringField()

        class DogOwner(db.Document):
            dogs = db.ListField(db.ReferenceField(Dog))

        DogOwnerForm = model_form(DogOwner, field_args={"dogs": {"allow_blank": True}})

        dogs = [Dog(name="fido"), Dog(name="rex")]
        for dog in dogs:
            dog.save()

        form = DogOwnerForm(dogs=dogs)
        assert form.validate()

        assert isinstance(form.dogs.widget, wtforms.widgets.Select)
        assert form.dogs.widget.multiple

        # Validate the options - both dogs should be selected and
        # there should be an extra blank option.
        choices = list(form.dogs)
        assert len(choices) == 3
        assert not choices[0].checked
        assert choices[0].data == "__None"
        assert choices[1].checked
        assert choices[1].data == dogs[0].pk
        assert choices[2].checked
        assert choices[2].data == dogs[1].pk

        # Validate selecting two items
        form = DogOwnerForm(MultiDict({"dogs": [dog.id for dog in dogs]}))
        assert form.dogs.data == dogs

        # Validate selecting none actually empties the list
        form = DogOwnerForm(MultiDict({"dogs": "__None"}), dogs=dogs)
        assert form.dogs.data is None


def test_modelselectfield_multiple_initalvalue_None(app, db):
    with app.test_request_context("/"):

        class Dog(db.Document):
            name = db.StringField()

        class DogOwner(db.Document):
            dogs = db.ListField(db.ReferenceField(Dog))

        DogOwnerForm = model_form(DogOwner)

        dogs = [Dog(name="fido"), Dog(name="rex")]
        for dog in dogs:
            dog.save()

        form = DogOwnerForm(dogs=None)
        assert form.validate()

        assert isinstance(form.dogs.widget, wtforms.widgets.Select)
        assert form.dogs.widget.multiple

        # Validate if both dogs are selected
        choices = list(form.dogs)
        assert len(choices) == 2
        assert not choices[0].checked
        assert not choices[1].checked


def test_modelradiofield(app, db):
    with app.test_request_context("/"):

        choices = [("male", "Male"), ("female", "Female"), ("other", "Other")]

        class Poll(db.Document):
            answer = db.StringField(choices=choices)

        PollForm = model_form(Poll, field_args={"answer": {"radio": True}})

        form = PollForm(answer=None)
        assert form.validate()

        assert form.answer.type == "RadioField"
        assert form.answer.choices == choices


def test_passwordfield(app, db):
    with app.test_request_context("/"):

        class User(db.Document):
            password = db.StringField()

        UserForm = model_form(User, field_args={"password": {"password": True}})
        form = UserForm(password="12345")
        assert isinstance(form.password.widget, wtforms.widgets.PasswordInput)


def test_unique_with(app, db):

    with app.test_request_context("/"):

        class Item(db.Document):
            owner_id = db.ObjectIdField(required=True)
            owner_item_id = db.StringField(required=True, unique_with="owner_id")

        Item.drop_collection()

        object_id = bson.ObjectId()
        Item(owner_id=object_id, owner_item_id="1").save()

        with pytest.raises(NotUniqueError):
            Item(owner_id=object_id, owner_item_id="1").save()

        assert 1 == Item.objects.count()


def test_sub_field_args(app, db):
    with app.test_request_context("/"):

        class TestModel(db.Document):
            lst = db.ListField(db.StringField())

        field_args = {
            "lst": {
                "label": "Custom Label",
                "field_args": {
                    "widget": wtforms.widgets.HiddenInput(),
                    "label": "Hidden Input",
                },
            }
        }
        CustomForm = model_form(TestModel, field_args=field_args)

        custom_form = CustomForm(obj=TestModel(lst=["Foo"]))
        list_label = flask.render_template_string(
            "{{ custom_form.lst.label }}", custom_form=custom_form
        )
        assert "Custom Label" in list_label
        assert "Hidden Input" not in list_label

        sub_label = flask.render_template_string(
            "{{ custom_form.lst }}", custom_form=custom_form
        )
        assert "Hidden Input" in sub_label


def test_modelselectfield_multiple_selected_elements_must_be_retained(app, db):
    with app.test_request_context("/"):

        class Dog(db.Document):
            name = db.StringField()

            def __unicode__(self):
                return self.name

        class DogOwner(db.Document):
            dogs = db.ListField(db.ReferenceField(Dog))

        DogOwnerForm = model_form(DogOwner)

        fido = Dog(name="fido").save()
        Dog(name="rex").save()

        dogOwner = DogOwner(dogs=[fido])
        form = DogOwnerForm(obj=dogOwner)
        html = form.dogs()

        m = re.search("<option selected .+?>(.*?)</option>", html)

        # Should have one selected option
        assert m is not None
        assert "fido" == m.group(1)


def test_model_form_help_text(app, db):
    with app.test_request_context("/"):

        class BlogPost(db.Document):
            title = db.StringField(
                required=True,
                help_text="Some imaginative title to set the world on fire",
            )

        post = BlogPost(title="hello world").save()

        BlogPostForm = model_form(BlogPost)
        form = BlogPostForm(instance=post)

        assert (
            form.title.description == "Some imaginative title to set the world on fire"
        )


def test_shared_field_args(app, db):
    with app.test_request_context("/"):

        class BlogPost(db.Document):
            title = db.StringField(required=True)
            content = db.StringField(required=False)

        shared_field_args = {
            "title": {"validators": [wtforms.validators.Regexp("test")]}
        }

        TitleOnlyForm = model_form(
            BlogPost, field_args=shared_field_args, exclude=["content"]
        )
        BlogPostForm = model_form(BlogPost, field_args=shared_field_args)

        # ensure shared field_args don't create duplicate validators
        title_only_form = TitleOnlyForm()
        assert len(title_only_form.title.validators) == 2

        blog_post_form = BlogPostForm()
        assert len(blog_post_form.title.validators) == 2


def test_embedded_model_form(app, db):
    with app.test_request_context("/"):

        class Content(db.EmbeddedDocument):
            text = db.StringField()
            lang = db.StringField(max_length=3)

        class Post(db.Document):
            title = db.StringField(max_length=120, required=True)
            tags = db.ListField(db.StringField(max_length=30))
            content = db.EmbeddedDocumentField("Content")

        PostForm = model_form(Post)
        form = PostForm()
        assert "content-text" in "%s" % form.content.text


def test_form_label_modifier(app, db):
    with app.test_request_context("/"):

        class FoodItem(db.Document):
            title = db.StringField()

        class FoodStore(db.Document):
            title = db.StringField(max_length=120, required=True)
            food_items = db.ListField(db.ReferenceField(FoodItem))

            def food_items_label_modifier(obj):
                return obj.title

        fruit_names = ["banana", "apple", "pear"]

        food_items = [FoodItem(title=name).save() for name in fruit_names]

        FoodStore(title="John's fruits", food_items=food_items).save()

        FoodStoreForm = model_form(FoodStore)
        form = FoodStoreForm()

        assert [obj.label.text for obj in form.food_items] == fruit_names



import inspect

import mongoengine
from flask import Flask, abort, current_app
from mongoengine.base.fields import BaseField
from mongoengine.errors import DoesNotExist
from mongoengine.queryset import QuerySet

from .connection import *
from .json import override_json_encoder
from .pagination import *
from .sessions import *
from .wtf import WtfBaseField

VERSION = (1, 0, 0)


def get_version():
    """Return the VERSION as a string."""
    return ".".join(map(str, VERSION))


__version__ = get_version()


def _patch_base_field(obj, name):
    """
    If the object submitted has a class whose base class is
    mongoengine.base.fields.BaseField, then monkey patch to
    replace it with flask_mongoengine.wtf.WtfBaseField.

    @note:  WtfBaseField is an instance of BaseField - but
            gives us the flexibility to extend field parameters
            and settings required of WTForm via model form generator.

    @see: flask_mongoengine.wtf.base.WtfBaseField.
    @see: model_form in flask_mongoengine.wtf.orm

    @param obj: MongoEngine instance in which we should locate the class.
    @param name: Name of an attribute which may or may not be a BaseField.
    """
    # TODO is there a less hacky way to accomplish the same level of
    # extensibility/control?

    # get an attribute of the MongoEngine class and return if it's not
    # a class
    cls = getattr(obj, name)
    if not inspect.isclass(cls):
        return

    # if it is a class, inspect all of its parent classes
    cls_bases = list(cls.__bases__)

    # if any of them is a BaseField, replace it with WtfBaseField
    for index, base in enumerate(cls_bases):
        if base == BaseField:
            cls_bases[index] = WtfBaseField
            cls.__bases__ = tuple(cls_bases)
            break

    # re-assign the class back to the MongoEngine instance
    delattr(obj, name)
    setattr(obj, name, cls)


def _include_mongoengine(obj):
    """
    Copy all of the attributes from mongoengine and mongoengine.fields
    onto obj (which should be an instance of the MongoEngine class).
    """
    # TODO why do we need this? What's wrong with importing from the
    # original modules?
    for module in (mongoengine, mongoengine.fields):
        for attr_name in module.__all__:
            if not hasattr(obj, attr_name):
                setattr(obj, attr_name, getattr(module, attr_name))

                # patch BaseField if available
                _patch_base_field(obj, attr_name)


def current_mongoengine_instance():
    """Return a MongoEngine instance associated with current Flask app."""
    me = current_app.extensions.get("mongoengine", {})
    for k, v in me.items():
        if isinstance(k, MongoEngine):
            return k


class MongoEngine(object):
    """Main class used for initialization of Flask-MongoEngine."""

    def __init__(self, app=None, config=None):
        _include_mongoengine(self)

        self.app = None
        self.config = config
        self.Document = Document
        self.DynamicDocument = DynamicDocument

        if app is not None:
            self.init_app(app, config)

    def init_app(self, app, config=None):
        if not app or not isinstance(app, Flask):
            raise TypeError("Invalid Flask application instance")

        self.app = app

        app.extensions = getattr(app, "extensions", {})

        # Make documents JSON serializable
        override_json_encoder(app)

        if "mongoengine" not in app.extensions:
            app.extensions["mongoengine"] = {}

        if self in app.extensions["mongoengine"]:
            # Raise an exception if extension already initialized as
            # potentially new configuration would not be loaded.
            raise ValueError("Extension already initialized")

        if config:
            # Passed config have max priority, over init config.
            self.config = config

        if not self.config:
            # If no configs passed, use app.config.
            config = app.config

        # Obtain db connection(s)
        connections = create_connections(config)

        # Store objects in application instance so that multiple apps do not
        # end up accessing the same objects.
        s = {"app": app, "conn": connections}
        app.extensions["mongoengine"][self] = s

    @property
    def connection(self):
        """
        Return MongoDB connection(s) associated with this MongoEngine
        instance.
        """
        return current_app.extensions["mongoengine"][self]["conn"]


class BaseQuerySet(QuerySet):
    """Mongoengine's queryset extended with handy extras."""

    def get_or_404(self, *args, **kwargs):
        """
        Get a document and raise a 404 Not Found error if it doesn't
        exist.
        """
        try:
            return self.get(*args, **kwargs)
        except DoesNotExist:
            message = kwargs.get("message", None)
            abort(404, message) if message else abort(404)

    def first_or_404(self, message=None):
        """Same as get_or_404, but uses .first, not .get."""
        obj = self.first()
        return obj if obj else abort(404, message) if message else abort(404)

    def paginate(self, page, per_page, **kwargs):
        """
        Paginate the QuerySet with a certain number of docs per page
        and return docs for a given page.
        """
        return Pagination(self, page, per_page)

    def paginate_field(self, field_name, doc_id, page, per_page, total=None):
        """
        Paginate items within a list field from one document in the
        QuerySet.
        """
        # TODO this doesn't sound useful at all - remove in next release?
        item = self.get(id=doc_id)
        count = getattr(item, field_name + "_count", "")
        total = total or count or len(getattr(item, field_name))
        return ListFieldPagination(
            self, doc_id, field_name, page, per_page, total=total
        )


class Document(mongoengine.Document):
    """Abstract document with extra helpers in the queryset class"""

    meta = {"abstract": True, "queryset_class": BaseQuerySet}

    def paginate_field(self, field_name, page, per_page, total=None):
        """Paginate items within a list field."""
        # TODO this doesn't sound useful at all - remove in next release?
        count = getattr(self, field_name + "_count", "")
        total = total or count or len(getattr(self, field_name))
        return ListFieldPagination(
            self.__class__.objects, self.pk, field_name, page, per_page, total=total
        )


class DynamicDocument(mongoengine.DynamicDocument):
    """Abstract Dynamic document with extra helpers in the queryset class"""

    meta = {"abstract": True, "queryset_class": BaseQuerySet}



import flask
import pytest

from flask_mongoengine import MongoEngine


@pytest.fixture()
def extended_db(app):
    app.json_encoder = DummyEncoder
    app.config["MONGODB_HOST"] = "mongodb://localhost:27017/flask_mongoengine_test_db"
    test_db = MongoEngine(app)
    db_name = test_db.connection.get_database("flask_mongoengine_test_db").name

    if not db_name.endswith("_test_db"):
        raise RuntimeError(
            f"DATABASE_URL must point to testing db, not to master db ({db_name})"
        )

    # Clear database before tests, for cases when some test failed before.
    test_db.connection.drop_database(db_name)

    yield test_db

    # Clear database after tests, for graceful exit.
    test_db.connection.drop_database(db_name)


class DummyEncoder(flask.json.JSONEncoder):
    """
    An example encoder which a user may create and override
    the apps json_encoder with.
    This class is a NO-OP, but used to test proper inheritance.
    """


@pytest.mark.usefixtures("extended_db")
def test_inheritance(app):
    assert issubclass(app.json_encoder, DummyEncoder)
    json_encoder_name = app.json_encoder.__name__

    # Since the class is dynamically derrived, must compare class names
    # rather than class objects.
    assert json_encoder_name == "MongoEngineJSONEncoder"



import mongoengine
from pymongo import ReadPreference, uri_parser


__all__ = (
    "create_connections",
    "get_connection_settings",
    "InvalidSettingsError",
)


MONGODB_CONF_VARS = (
    "MONGODB_ALIAS",
    "MONGODB_DB",
    "MONGODB_HOST",
    "MONGODB_IS_MOCK",
    "MONGODB_PASSWORD",
    "MONGODB_PORT",
    "MONGODB_USERNAME",
    "MONGODB_CONNECT",
    "MONGODB_TZ_AWARE",
)


class InvalidSettingsError(Exception):
    pass


def _sanitize_settings(settings):
    """Given a dict of connection settings, sanitize the keys and fall
    back to some sane defaults.
    """
    # Remove the "MONGODB_" prefix and make all settings keys lower case.
    resolved_settings = {}
    for k, v in settings.items():
        if k.startswith("MONGODB_"):
            k = k[len("MONGODB_") :]
        k = k.lower()
        resolved_settings[k] = v

    # Handle uri style connections
    if "://" in resolved_settings.get("host", ""):
        # this section pulls the database name from the URI
        # PyMongo requires URI to start with mongodb:// to parse
        # this workaround allows mongomock to work
        uri_to_check = resolved_settings["host"]

        if uri_to_check.startswith("mongomock://"):
            uri_to_check = uri_to_check.replace("mongomock://", "mongodb://")

        uri_dict = uri_parser.parse_uri(uri_to_check)
        resolved_settings["db"] = uri_dict["database"]

    # Add a default name param or use the "db" key if exists
    if resolved_settings.get("db"):
        resolved_settings["name"] = resolved_settings.pop("db")
    else:
        resolved_settings["name"] = "test"

    # Add various default values.
    resolved_settings["alias"] = resolved_settings.get(
        "alias", mongoengine.DEFAULT_CONNECTION_NAME
    )  # TODO do we have to specify it here? MongoEngine should take care of that
    resolved_settings["host"] = resolved_settings.get(
        "host", "localhost"
    )  # TODO this is the default host in pymongo.mongo_client.MongoClient, we may not need to explicitly set a default here
    resolved_settings["port"] = resolved_settings.get(
        "port", 27017
    )  # TODO this is the default port in pymongo.mongo_client.MongoClient, we may not need to explicitly set a default here

    # Default to ReadPreference.PRIMARY if no read_preference is supplied
    resolved_settings["read_preference"] = resolved_settings.get(
        "read_preference", ReadPreference.PRIMARY
    )

    # Clean up empty values
    for k, v in list(resolved_settings.items()):
        if v is None:
            del resolved_settings[k]

    return resolved_settings


def get_connection_settings(config):
    """
    Given a config dict, return a sanitized dict of MongoDB connection
    settings that we can then use to establish connections. For new
    applications, settings should exist in a "MONGODB_SETTINGS" key, but
    for backward compactibility we also support several config keys
    prefixed by "MONGODB_", e.g. "MONGODB_HOST", "MONGODB_PORT", etc.
    """
    # Sanitize all the settings living under a "MONGODB_SETTINGS" config var
    if "MONGODB_SETTINGS" in config:
        settings = config["MONGODB_SETTINGS"]

        # If MONGODB_SETTINGS is a list of settings dicts, sanitize each
        # dict separately.
        if isinstance(settings, list):
            # List of connection settings.
            settings_list = []
            for setting in settings:
                settings_list.append(_sanitize_settings(setting))
            return settings_list

        # Otherwise, it should be a single dict describing a single connection.
        else:
            return _sanitize_settings(settings)

    # If "MONGODB_SETTINGS" doesn't exist, sanitize the "MONGODB_" keys
    # as if they all describe a single connection.
    else:
        config = dict(
            (k, v) for k, v in config.items() if k in MONGODB_CONF_VARS
        )  # ugly dict comprehention in order to support python 2.6
        return _sanitize_settings(config)


def create_connections(config):
    """
    Given Flask application's config dict, extract relevant config vars
    out of it and establish MongoEngine connection(s) based on them.
    """
    # Validate that the config is a dict
    if config is None or not isinstance(config, dict):
        raise InvalidSettingsError("Invalid application configuration")

    # Get sanitized connection settings based on the config
    conn_settings = get_connection_settings(config)

    # If conn_settings is a list, set up each item as a separate connection
    # and return a dict of connection aliases and their connections.
    if isinstance(conn_settings, list):
        connections = {}
        for each in conn_settings:
            alias = each["alias"]
            connections[alias] = _connect(each)
        return connections

    # Otherwise, return a single connection
    return _connect(conn_settings)


def _connect(conn_settings):
    """Given a dict of connection settings, create a connection to
    MongoDB by calling mongoengine.connect and return its result.
    """
    db_name = conn_settings.pop("name")
    return mongoengine.connect(db_name, **conn_settings)



import os
from setuptools import setup


description = (
    "Flask-MongoEngine is a Flask extension "
    "that provides integration with MongoEngine and WTF model forms."
)

# Load index.rst as long_description
doc_path = os.path.join(os.path.dirname(__file__), "docs", "index.rst")
long_description = open(doc_path, encoding="utf-8").read()
long_description_content_type = "text/x-rst"

# Stops exit traceback on tests
try:
    import multiprocessing  # noqa
except ImportError:
    pass


def get_version(version_tuple):
    """Return the version tuple as a string, e.g. for (0, 10, 7),
    return '0.10.7'.
    """
    return ".".join(map(str, version_tuple))


# Dirty hack to get version number from flask_monogengine/__init__.py - we
# can't import it as it depends on PyMongo and PyMongo isn't installed until
# this file is read
init = os.path.join(os.path.dirname(__file__), "flask_mongoengine", "__init__.py")
version_line = list(filter(lambda l: l.startswith("VERSION"), open(init)))[0]
version = get_version(eval(version_line.split("=")[-1]))

test_requirements = ["coverage", "pytest", "pytest-cov"]

setup(
    name="flask-mongoengine",
    version=version,
    url="https://github.com/mongoengine/flask-mongoengine",
    license="BSD",
    author="Ross Lawley",
    author_email="ross.lawley@gmail.com",
    zip_safe=False,
    platforms="any",
    install_requires=[
        "Flask>=1.1.2",
        "WTForms[email]>=2.3.1",
        "Flask-WTF>=0.14.3",
        "mongoengine>=0.20",
    ],
    packages=["flask_mongoengine", "flask_mongoengine.wtf"],
    include_package_data=True,
    tests_require=test_requirements,
    description=description,
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Flask",
    ],
)



import flask

from .models import Todo


def index():
    # As a list to test debug toolbar
    Todo.objects().delete()  # Removes
    Todo(title="Simple todo A", text="12345678910").save()  # Insert
    Todo(title="Simple todo B", text="12345678910").save()  # Insert
    Todo.objects(title__contains="B").update(set__text="Hello world")  # Update
    todos = list(Todo.objects[:10])
    todos = Todo.objects.all()
    return flask.render_template("index.html", todos=todos)


def pagination():
    Todo.objects().delete()
    for i in range(10):
        Todo(title="Simple todo {}".format(i), text="12345678910").save()  # Insert

    page_num = int(flask.request.args.get("page") or 1)
    todos_page = Todo.objects.paginate(page=page_num, per_page=3)

    return flask.render_template("pagination.html", todos_page=todos_page)
